div.admonitionblock .content {
    border-left: 2px solid silver;
    padding-left: 0.5em;
}

div.admonitionblock .icon {
    vertical-align: top;
    padding-right: 0.5em;
}

dd {
    padding-left: 2em;
}

p {
    text-align: justify;
}

.footer {
    border-top: 2pt solid silver;
    margin-top: 20px;
    padding-top: 10px;
}

.rfloat {
    float: right;
    margin-left: 0.5em;
}

.frame {
    background-color: #f9f9f9;
    border: 1px solid #cccccc;
    padding: 4px;
    margin-top: 5px;
}